import copy
import re
import types

import numpy as np
import pandas as pd

from seeq.sdk import *
from seeq.spy import _common, _login, _metadata
from seeq.spy._errors import *
from seeq.spy._session import Session
from seeq.spy.assets._trees import _constants, _match, _path, _utils
from seeq.spy.workbooks._item import Item


def apply_friendly_name(df) -> None:
    if 'Friendly Name' not in df.columns or df['Friendly Name'].isnull().all():
        _common.put_properties_on_df(df, types.SimpleNamespace(modified_items=set()))
        return

    # If we are changing the names of items in a dataframe whose paths are dependent on one another, then
    # record those dependencies so we can modify paths afterwards as well
    relationships = path_relationships(df)

    modified_items = set()
    for i in df.index:
        if pd.isnull(df.loc[i, 'Friendly Name']):
            continue
        if _match.is_column_value_query(df.loc[i, 'Friendly Name']):
            new_name = _match.fill_column_values(df.loc[i], df.loc[i, 'Friendly Name'])
        else:
            new_name = df.loc[i, 'Friendly Name']
        if pd.isnull(new_name):
            continue
        df.loc[i, 'Name'] = new_name
        if _common.present(df.loc[i], 'ID'):
            modified_items.add(df.loc[i, 'ID'])

    recover_relationships(df, relationships)
    _common.put_properties_on_df(df, types.SimpleNamespace(modified_items=modified_items))


def path_relationships(df) -> dict:
    """
    Return a dict of dicts indicating via integers how the paths of the input rows are dependent on one another.

    Example:
        df = pd.DataFrame([{
            'Path': 'Root', 'Name': 'Item 1'
        }, {
            'Path': 'Root >> Item 1', 'Name': 'Item 2'
        }])

        output = {
            1: { # 1 refers here to the row in df with index 1, i.e., Item 2
                1: 0 # 1 refers here to the item in Item 2's path with index 1, i.e. 'Item 1'
                     # 0 refers here to the index of Item 1's row in df
            }
        }
    """
    if len(df) == 0 or 'Name' not in df.columns:
        return None
    temp_df = df[['Name']].copy()
    temp_df['Path'] = df.apply(_path.determine_path, axis=1)
    full_paths = list(temp_df.apply(_path.get_full_path, axis=1).apply(_common.path_string_to_list))
    relationships = dict()
    # This is O(n^2) but it's not a core feature
    for i, this in enumerate(full_paths):
        if this == [''] or len(this) == 0:
            continue
        for j, other in enumerate(full_paths):
            if other == ['']:
                continue
            # If the full path "other" begins with the full path "this", then we mark that
            # the (len(this) - 1)th element in "other"'s path is equal to "this"
            if len(other) > len(this) and other[:len(this)] == this:
                if j not in relationships:
                    relationships[j] = dict()
                relationships[j][len(this) - 1] = i
    return relationships


def recover_relationships(df, relationships) -> None:
    """
    Takes a list of relationships (in the format described in _path_relationships) and modifies paths in
    df to reflect those relationships
    """
    if relationships is None:
        return
    for i, path_ref_dict in relationships.items():
        path = _path.determine_path(df.loc[i])
        path_list = _common.path_string_to_list(path) if path else []
        for j, reference in path_ref_dict.items():
            if 0 <= reference < len(df) and 0 <= j < len(path_list):
                path_list[j] = df.loc[reference, 'Name']
        df.loc[i, 'Path'] = _common.path_list_to_string(path_list)
    if 'Asset' in df.columns:
        df.drop(columns='Asset', inplace=True)


def process_properties(session: Session, df, status, existing_tree_df=None, pull_nodes=True,
                       keep_parent_column=False) -> pd.DataFrame:
    """
    Sanitize and pull item properties into an input dataframe. Steps in order:
    -- Pulls missing properties for items with ID provided
    -- Filters out properties not in _constants.dataframe_columns
    -- Determines tree depth
    -- Determines (if possible_tree_copy is True) if the input dataframe corresponds to an existing SPy tree
        -- If it is indeed a copy of a SPy tree, pulls in calculations from the original tree
        -- Otherwise, it converts all items with IDs into references
    -- Ensures all formula parameters are NAN or dict
    """
    df = df.reset_index(drop=True)

    df = df.apply(process_row_properties, axis=1,
                  session=session,
                  status=status,
                  pull_nodes=pull_nodes,
                  keep_parent_column=keep_parent_column)

    def _row_is_from_existing_tree(row):
        if existing_tree_df is None or not _common.present(row, 'ID'):
            return 'new'
        same_id_rows = existing_tree_df[existing_tree_df.ID.str.casefold() == row['ID'].casefold()]
        if len(same_id_rows) != 1:
            return 'new'
        if _common.present(row, 'Type') and row['Type'].casefold() != same_id_rows.Type.iloc[0].casefold():
            return 'new'
        if _common.present(row, 'Name') and row['Name'].casefold() != same_id_rows.Name.iloc[0].casefold():
            return 'modified'
        if _common.present(row, 'Path') and row['Path'].casefold() != same_id_rows.Path.iloc[0].casefold():
            return 'modified'
        return 'pre-existing'

    row_type = df.apply(_row_is_from_existing_tree, axis=1)
    modified_items = df.loc[row_type == 'modified', 'ID'] if 'ID' in df.columns else set()

    # For the nodes that originated from the pre-existing SPy tree we are modifying, we want to pull
    # pre-existing calculations directly.
    formulas_api = FormulasApi(session.client)
    metrics_api = MetricsApi(session.client)
    df.loc[row_type == 'pre-existing', :] = df.loc[row_type == 'pre-existing', :] \
        .apply(pull_calculation, axis=1, formulas_api=formulas_api, metrics_api=metrics_api)

    # For the nodes that originate from places other than the pre-existing SPy tree we are modifying,
    # we want to build references so we create and modify *copies* and not the original items.
    df.loc[row_type != 'pre-existing', :] = df.loc[row_type != 'pre-existing', :] \
        .apply(make_node_reference, axis=1, session=session, formulas_api=formulas_api, metrics_api=metrics_api)

    if 'Formula Parameters' in df.columns:
        df['Formula Parameters'] = df['Formula Parameters'].apply(formula_parameters_to_dict)

    _common.put_properties_on_df(df, types.SimpleNamespace(modified_items=modified_items))

    return df


# Note that the session argument is second in this special case because we call this from pd.DataFrame.apply(),
# which requires that row is the first argument.
def process_row_properties(row, session: Session, status, pull_nodes, keep_parent_column) -> pd.Series:
    if _common.present(row, 'ID') and pull_nodes:
        new_row = pull_node(session, row['ID'])
        _utils.increment_status_df(status, pulled_items=[new_row])
    else:
        new_row = pd.Series(index=_constants.dataframe_columns, dtype='object')

    # In case that properties are specified, but IDs are given, the user-given properties
    # override those pulled from Seeq
    for prop, value in row.items():
        if prop == 'Type' and _common.present(new_row, 'Type'):
            continue
        if prop in ['Path', 'Asset']:
            prop = 'Path'
            value = _path.determine_path(row)
        add_tree_property(new_row, prop, value)

    if not _common.present(new_row, 'Type') and not _common.present(new_row, 'Formula'):
        if _common.present(new_row, 'Measured Item'):
            new_row['Type'] = 'Metric'
        else:
            new_row['Type'] = 'Asset'

    if not _common.present(new_row, 'Path'):
        new_row['Path'] = ''
    new_row['Depth'] = new_row['Path'].count('>>') + 2 if new_row['Path'] else 1

    if keep_parent_column and _common.present(row, 'Parent'):
        new_row['Parent'] = row['Parent']

    return new_row


def make_node_reference(row, session: Session, formulas_api: FormulasApi, metrics_api: MetricsApi) -> pd.Series:
    row = row.copy()
    if _common.present(row, 'ID'):
        if _common.get(row, 'Type') in _constants.data_types and not is_reference(row):
            _metadata.build_reference(session, row)
        elif 'Metric' in _common.get(row, 'Type'):
            row = pull_calculation(row, formulas_api, metrics_api)
        if _common.present(row, 'ID'):
            row['Referenced ID'] = row['ID']
    row['ID'] = np.nan
    return row


def is_reference(row) -> bool:
    if not _common.get(row, 'Referenced ID') or not _common.get(row, 'Formula Parameters'):
        return False
    formula = _common.get(row, 'Formula')
    if formula is not None and re.match(r'^\$\w+$', formula):
        return True
    else:
        return False


def pull_calculation(row, formulas_api: FormulasApi, metrics_api: MetricsApi) -> pd.Series:
    if _common.get(row, 'Type') in _constants.calculated_types and _common.present(row, 'ID'):
        row = row.copy()
        formula_output = formulas_api.get_item(id=row['ID'])  # type: FormulaItemOutputV1
        row['Formula'] = formula_output.formula
        row['Formula Parameters'] = [
            '%s=%s' % (p.name, p.item.id if p.item else p.formula) for p in formula_output.parameters
        ]
    elif 'Metric' in _common.get(row, 'Type') and _common.present(row, 'ID'):
        row = row.copy()
        row['Type'] = 'Metric'

        def str_from_scalar_value_output(scalar_value_output):
            # noinspection PyProtectedMember
            return Item._str_from_scalar_value_dict(Item._dict_from_scalar_value_output(scalar_value_output)).strip()

        def set_metric_value_if_present(column_name, val):
            if val is not None:
                if isinstance(val, ScalarValueOutputV1):
                    row[column_name] = str_from_scalar_value_output(val)
                elif isinstance(val, ItemPreviewWithAssetsV1):
                    row[column_name] = val.id
                else:
                    row[column_name] = val

        metric = metrics_api.get_metric(id=row['ID'])  # type: ThresholdMetricOutputV1
        set_metric_value_if_present('Aggregation Function', metric.aggregation_function)
        set_metric_value_if_present('Bounding Condition', metric.bounding_condition)
        set_metric_value_if_present('Bounding Condition Maximum Duration', metric.bounding_condition_maximum_duration)
        set_metric_value_if_present('Duration', metric.duration)
        set_metric_value_if_present('Measured Item', metric.measured_item)
        set_metric_value_if_present('Metric Neutral Color', metric.neutral_color)
        set_metric_value_if_present('Number Format', metric.number_format)
        set_metric_value_if_present('Period', metric.period)
        set_metric_value_if_present('Process Type', metric.process_type)
        if metric.thresholds:
            thresholds_dict = dict()
            for threshold in metric.thresholds:  # type: ThresholdOutputV1
                # Key can be 'HiHi#FF0000' or 'HiHi'
                key = threshold.priority.name
                if threshold.priority.color:
                    key += threshold.priority.color
                # Value could get pulled as an ID or as a string-ified value
                value = ''
                if not threshold.is_generated and threshold.item:
                    value = threshold.item.id
                elif threshold.value is not None:
                    if isinstance(threshold.value, ScalarValueOutputV1):
                        value = str_from_scalar_value_output(threshold.value)
                    else:
                        value = threshold.value
                if key and value:
                    thresholds_dict[key] = value
            row['Thresholds'] = thresholds_dict

    return row


def pull_node(session: Session, node_id) -> pd.Series:
    """
    Returns a dataframe row corresponding to the item given by node_id
    """
    items_api = _login.get_api(session, ItemsApi)

    item_output = items_api.get_item_and_all_properties(id=node_id)  # type: ItemOutputV1
    node = pd.Series(index=_constants.dataframe_columns, dtype='object')

    # Extract only the properties we use
    node['Name'] = item_output.name
    node['Type'] = item_output.type
    node['ID'] = item_output.id  # If this should be a copy, it'll be converted to 'Referenced ID' later
    for prop in item_output.properties:  # type: PropertyOutputV1
        add_tree_property(node, prop.name, prop.value)
    if 'Metric' in node['Type']:
        node = pull_calculation(node, _login.get_api(session, FormulasApi), _login.get_api(session, MetricsApi))

    return node


def add_tree_property(properties, key, value) -> pd.Series:
    """
    If the property is one which is used by SPy Trees, adds the key+value pair to the dict.
    """
    if key in _constants.dataframe_columns:
        value = _common.none_to_nan(value)
        if isinstance(value, str) and key in ['Cache Enabled', 'Archived', 'Enabled', 'Unsearchable']:
            # Ensure that these are booleans. Otherwise Seeq Server will silently ignore them.
            value = (value.lower() == 'true')
        if key not in properties or not (pd.api.types.is_scalar(value) and pd.isnull(value)):
            properties[key] = value
    return properties


def formula_parameters_to_dict(formula_parameters) -> dict:
    if isinstance(formula_parameters, dict) or (pd.api.types.is_scalar(formula_parameters) and pd.isnull(
            formula_parameters)):
        return formula_parameters

    if isinstance(formula_parameters, str):  # formula_parameters == 'x=2b17adfd-3308-4c03-bdfb-bf4419bf7b3a'
        # handle an empty string case
        if len(formula_parameters) == 0:
            return dict()
        else:
            formula_parameters = [formula_parameters]

    if isinstance(formula_parameters, pd.Series):
        formula_parameters = formula_parameters.tolist()

    formula_dictionary = dict()
    if isinstance(formula_parameters, list):  # formula_parameters == ['x=2b17adfd-3308-4c03-bdfb-bf4419bf7b3a', ...]
        for param in formula_parameters:  # type: str
            split_list = param.split('=')  # ['x', '2b17...']
            if len(split_list) != 2:
                raise SPyException(f'Formula Parameter: {param} needs to be in the format \'paramName=inputItem\'.')
            formula_dictionary[split_list[0].strip()] = split_list[1].strip()
    return formula_dictionary  # output == {'x': '2b17adfd-3308-4c03-bdfb-bf4419bf7b3a'}


def format_formula_parameters(df, status) -> pd.DataFrame:
    """
    For each applicable row in the dataframe, resolve formula parameters from relative path to full path. If the
    referred item does not exist in the dataframe, error.
    :param df: The internal _dataframe to transform for _metadata.push().
    :param status: The status object if an error is encountered.
    :return: A deep copy of the original dataframe, but with fully resolved formula parameters.
    """
    output_df = df.copy()

    output_formula_parameters_column = pd.Series(np.nan, index=output_df.index, dtype='object')

    # Takes relative-path formula parameters and changes them to full path for the ensuing push call
    for idx, row in output_df[output_df['Formula Parameters'].notnull()].iterrows():
        formula_parameters = copy.deepcopy(row['Formula Parameters'])

        for name, item in row['Formula Parameters'].items():
            if not isinstance(item, str) or _common.is_guid(item):
                continue
            item_full_path = _path.get_full_path({'Path': row['Path'], 'Name': item})

            resolved_path, _ = resolve_node_reference(row, item, output_df)
            if resolved_path is None:
                # Validation prevents this error from being raised
                e = SPyValueError(f"Issue resolving formula parameters for item at '{row.Path} >> "
                                  f"{row.Name}'. No matches were found for '{item_full_path}'.")
                status.exception(e, throw=True)
            formula_parameters[name] = resolved_path

        output_formula_parameters_column[idx] = formula_parameters

    output_df['Formula Parameters'] = output_formula_parameters_column
    return output_df


def format_metric_parameters(df) -> pd.DataFrame:
    """
    For each applicable row in the dataframe, resolve the Measured Item and Bounding Condition from relative path
    to full path. Also converts Threshold values from relative path to full row dict.
    :param df: The internal _dataframe to transform for _metadata.push().
    :return: A deep copy of the original dataframe, but with fully resolved Metric references.
    """
    output_df = df.copy()

    output_measured_items_column = pd.Series(np.nan, index=output_df.index, dtype='object')
    output_bounding_conditions_column = pd.Series(np.nan, index=output_df.index, dtype='object')
    output_thresholds_column = pd.Series(np.nan, index=output_df.index, dtype='object')

    for idx, row in output_df[output_df['Type'].str.contains('Metric').fillna(False)].iterrows():
        if not pd.isnull(row['Measured Item']):
            resolved_path, _ = resolve_node_reference(row, row['Measured Item'], output_df)
            if resolved_path:
                output_measured_items_column[idx] = resolved_path
            else:
                output_measured_items_column[idx] = row['Measured Item']

        if not pd.isnull(row['Bounding Condition']):
            resolved_path, _ = resolve_node_reference(row, row['Bounding Condition'], output_df)
            if resolved_path:
                output_bounding_conditions_column[idx] = resolved_path
            else:
                output_bounding_conditions_column[idx] = row['Bounding Condition']

        if not pd.isnull(row['Thresholds']):
            thresholds = copy.deepcopy(row['Thresholds'])

            for key, threshold in row['Thresholds'].items():
                if not isinstance(threshold, str) or _common.is_guid(threshold):
                    continue
                resolved_path, resolved_row = resolve_node_reference(row, threshold, output_df)
                if resolved_path:
                    # Use a dict so _push can differentiate a string Threshold and a reference to another item
                    thresholds[key] = {
                        'Name': resolved_row['Name'],
                        'Path': resolved_row['Path']
                    }

            output_thresholds_column[idx] = thresholds

    output_df['Measured Item'] = output_measured_items_column
    output_df['Bounding Condition'] = output_bounding_conditions_column
    output_df['Thresholds'] = output_thresholds_column
    return output_df


def resolve_node_reference(current_row, referenced_item, df) -> (str, pd.Series):
    """
    Finds the item in the tree that is referenced as an input parameter by current_row.
    E.G. if a calculation 'Cooling Tower >> Area A >> Too Hot' references 'Sensors >> Temperature',
    the 'Cooling Tower >> Area A >> Sensors >> Temperature' row will be returned if it exists.

    :param current_row: The item row being evaluated that contains the reference.
    :param referenced_item: The name or path+name of the item being referenced. Must be at or below the current_row
    asset level.
    :param df: The entire tree dataframe
    :return: The full path string of the referenced item and its row, or None if not found.
    """
    item_full_path = _path.get_full_path({'Path': current_row['Path'], 'Name': referenced_item})
    for _, other_row in df.iterrows():
        if other_row is current_row:
            continue
        if _match.is_node_match(item_full_path, other_row):
            return _path.get_full_path(other_row), other_row
    return None, None



import re
from dataclasses import dataclass
from enum import Enum
from urllib.parse import urlparse, ParseResult

from seeq.spy import _common
from seeq.spy._errors import *


class HttpProtocol(Enum):
    HTTP = 1
    HTTPS = 2


class Route(Enum):
    HOME = 1
    WORKBOOK_EDIT = 2
    WORKBOOK_VIEW = 3
    WORKBOOK_PRESENTATION = 4
    JOURNAL_WORKSTEP_LINK = 5
    DATALAB_PROJECT = 6
    UNKNOWN = 7


@dataclass
class SeeqURL:
    protocol: HttpProtocol
    hostname: str
    port: int
    route: Route
    _folder_id: str
    _workbook_id: str
    _worksheet_id: str
    _workstep_id: str
    _datalab_project_id: str

    def __init__(self, *, protocol=None, hostname=None, port=None, route=None, folder_id=None, workbook_id=None,
                 worksheet_id=None, workstep_id=None, datalab_project_id=None):
        self.protocol = protocol
        self.hostname = hostname
        self.port = port
        self.route = route
        self.folder_id = folder_id
        self.workbook_id = workbook_id
        self.worksheet_id = worksheet_id
        self.workstep_id = workstep_id
        self.datalab_project_id = datalab_project_id

    @property
    def folder_id(self):
        return self._folder_id

    @folder_id.setter
    def folder_id(self, val):
        self._folder_id = None if val is None else val.upper()

    @property
    def workbook_id(self):
        return self._workbook_id

    @workbook_id.setter
    def workbook_id(self, val):
        self._workbook_id = None if val is None else val.upper()

    @property
    def worksheet_id(self):
        return self._worksheet_id

    @worksheet_id.setter
    def worksheet_id(self, val):
        self._worksheet_id = None if val is None else val.upper()

    @property
    def workstep_id(self):
        return self._workstep_id

    @workstep_id.setter
    def workstep_id(self, val):
        self._workstep_id = None if val is None else val.upper()

    @property
    def datalab_project_id(self):
        return self._datalab_project_id

    @datalab_project_id.setter
    def datalab_project_id(self, val):
        self._datalab_project_id = None if val is None else val.upper()

    @property
    def address(self):
        address = f"{'http' if self.protocol == HttpProtocol.HTTP else 'https'}://{self.hostname}"
        if (self.protocol == HttpProtocol.HTTP and self.port != 80) or \
                (self.protocol == HttpProtocol.HTTPS and self.port != 443):
            address += f':{self.port}'
        return address

    @property
    def path(self):
        if self.route == Route.WORKBOOK_EDIT:
            if self.folder_id is not None:
                return f'/{self.folder_id}/workbook/{self.workbook_id}/worksheet/{self.worksheet_id}'
            else:
                return f'/workbook/{self.workbook_id}/worksheet/{self.worksheet_id}'
        elif self.route == Route.WORKBOOK_VIEW:
            return f'/view/{self.worksheet_id}'
        elif self.route == Route.WORKBOOK_PRESENTATION:
            return f'/present/worksheet/{self.workbook_id}/{self.worksheet_id}'
        elif self.route == Route.JOURNAL_WORKSTEP_LINK:
            return f'/links'
        elif self.route == Route.HOME:
            if self.folder_id is not None:
                return f'/{self.folder_id}/folder/'
            else:
                return f'/workbooks'
        elif self.route == Route.DATALAB_PROJECT:
            return f'/data-lab/{self.datalab_project_id}'

        return ''

    @property
    def parameters(self):
        if self.route == Route.WORKBOOK_VIEW:
            if self.workstep_id is not None:
                return f'?workstepId={self.workstep_id}'
        elif self.route == Route.JOURNAL_WORKSTEP_LINK:
            return f'?type=workstep&workbook={self.workbook_id}' \
                   f'&worksheet={self.worksheet_id}&workstep={self.workstep_id}'

        return ''

    @property
    def url(self):
        return f'{self.address}{self.path}{self.parameters}'

    ADDRESS_REGEX = re.compile(r'^(?P<protocol>https?)://(?P<hostname>[^:/?]+)(?::(?P<port>\d+))?', re.IGNORECASE)

    ROUTES = {
        Route.WORKBOOK_EDIT: (re.compile(rf'^(?:/(?P<folder_id>{_common.GUID_REGEX}))?'
                                         rf'/workbook/(?P<workbook_id>{_common.GUID_REGEX})'
                                         rf'/worksheet/(?P<worksheet_id>{_common.GUID_REGEX})', re.IGNORECASE),
                              dict()),
        Route.WORKBOOK_VIEW: (re.compile(rf'^/view/(?P<worksheet_id>{_common.GUID_REGEX})', re.IGNORECASE),
                              {'workstepId': 'workstep_id'}),
        Route.WORKBOOK_PRESENTATION: (re.compile(rf'^/present/worksheet/(?P<workbook_id>{_common.GUID_REGEX})'
                                                 rf'/(?P<worksheet_id>{_common.GUID_REGEX})', re.IGNORECASE),
                                      dict()),
        Route.JOURNAL_WORKSTEP_LINK: (re.compile(rf'^/links', re.IGNORECASE), {'type': 'link_type',
                                                                               'workbook': 'workbook_id',
                                                                               'worksheet': 'worksheet_id',
                                                                               'workstep': 'workstep_id'}),
        Route.HOME: (re.compile(rf'^/(workbooks|(?P<folder_id>{_common.GUID_REGEX})/folder)', re.IGNORECASE),
                     dict()),
        Route.DATALAB_PROJECT: (
            re.compile(rf'/data-lab/(?P<datalab_project_id>{_common.GUID_REGEX})', re.IGNORECASE), dict())
    }

    @staticmethod
    def parse(url):
        address_match = SeeqURL.ADDRESS_REGEX.search(url)
        if not address_match:
            raise SPyRuntimeError(f'URL "{url}" does not start with typical http(s)://hostname(:port)')

        seeq_url = SeeqURL()
        groups = address_match.groupdict()
        seeq_url.protocol = HttpProtocol[groups['protocol'].upper()]
        seeq_url.hostname = groups['hostname']
        seeq_url.port = 80 if seeq_url.protocol == HttpProtocol.HTTP else 443
        if _common.present(groups, 'port'):
            seeq_url.port = int(groups['port'])

        path_and_parameters = url[address_match.end():].split('?')
        path = path_and_parameters[0]
        parameters = ''.join(path_and_parameters[1:]) if len(path_and_parameters) > 1 else None

        parameter_dict = dict()
        if parameters is not None:
            parameter_list = re.split(_common.HTML_AMPERSAND_REGEX, parameters)
            for parameter in parameter_list:
                parameter_parts = re.split(_common.HTML_EQUALS_REGEX, parameter)
                parameter_key = parameter_parts[0]
                parameter_value = parameter_parts[1] if len(parameter_parts) > 1 else None
                parameter_dict[parameter_key] = parameter_value

        seeq_url.route = Route.UNKNOWN
        for route in SeeqURL.ROUTES.keys():
            route_path_regex, route_parameters_dict = SeeqURL.ROUTES[route]
            route_match = route_path_regex.search(path)
            if not route_match:
                continue

            seeq_url.route = route
            groups = dict(route_match.groupdict())
            for parameter_key, parameter_value in parameter_dict.items():
                if parameter_key in route_parameters_dict:
                    groups[route_parameters_dict[parameter_key]] = parameter_value

            if route == Route.JOURNAL_WORKSTEP_LINK and groups['link_type'] != 'workstep':
                raise SPyValueError(f'Unknown JOURNAL_WORKSTEP_LINK type "{groups["link_type"]}"')

            seeq_url.folder_id = _common.get(groups, 'folder_id')
            seeq_url.workbook_id = _common.get(groups, 'workbook_id')
            seeq_url.worksheet_id = _common.get(groups, 'worksheet_id')
            seeq_url.workstep_id = _common.get(groups, 'workstep_id')
            seeq_url.datalab_project_id = _common.get(groups, 'datalab_project_id')

            break

        return seeq_url


def cleanse_url(url):
    parts = urlparse(url)  # type: ParseResult
    url = f'{parts.scheme}://{parts.hostname}'
    if parts.port is not None:
        if parts.port not in [80, 443] or \
                (parts.scheme.lower() == 'http' and parts.port != 80) or \
                (parts.scheme.lower() == 'https' and parts.port != 443):
            url += f':{parts.port}'

    return url


def parse_url(url):
    return SeeqURL.parse(url)


def get_workbook_id_from_url(url):
    """
    Get the Seeq ID of a workbook from a URL

    Given a URL copied from a browser or from an API response 'href' attribute,
    get the URL of the workbook.

    Parameters
    ----------
    url : str
        The URL

    Returns
    -------
    {str, None}
        The Seeq ID as a string, or None if no workbook ID was found
    """
    return SeeqURL.parse(url).workbook_id


def get_worksheet_id_from_url(url):
    """
    Get the Seeq ID of a worksheet from a URL

    Given a URL copied from a browser or from an API response 'href' attribute,
    get the URL of the worksheet.

    Parameters
    ----------
    url : str
        The URL

    Returns
    -------
    {str, None}
        The Seeq ID as a string, or None if no worksheet ID was found
    """
    return SeeqURL.parse(url).worksheet_id


def get_workstep_id_from_url(url):
    """
    Get the Seeq ID of a workstep from a URL

    Given a URL copied from a browser or from an API response 'href' attribute,
    get the URL of the workstep.

    Note that URLs from a browser rarely specify the workstep. URLs that
    contain the workstep normally come from links in Organizer Topic Documents
    or href attributes in API outputs.

    Parameters
    ----------
    url : str
        The URL

    Returns
    -------
    {str, None}
        The Seeq ID as a string, or None if no workstep ID was found
    """
    return SeeqURL.parse(url).workstep_id


def get_data_lab_project_id_from_url(url):
    """
    Get the Seeq DataLab Project ID from a URL

    Given a URL copied from a browser or from an API response 'href' attribute,
    get the URL of the DataLab project.

    Parameters
    ----------
    url : str
        The URL

    Returns
    -------
    {str, None}
        The Seeq ID as a string, or None if no project ID was found
    """
    return SeeqURL.parse(url).datalab_project_id



import pandas as pd

from seeq.spy import _common, _push
from seeq.spy.assets._trees import _utils


def get_full_path(node):
    if not isinstance(_common.get(node, 'Name'), str) or len(node['Name']) == 0:
        return ''
    if isinstance(_common.get(node, 'Path'), str) and len(node['Path']) != 0:
        return f"{node['Path']} >> {node['Name']}"
    return node['Name']


def sort_by_node_path(df):
    decorate_with_full_path(df)
    df.sort_values(by='Full Path List', inplace=True, ignore_index=True)
    remove_full_path(df)


def decorate_with_full_path(df):
    """
    From the 'Path' and 'Name' columns, add a 'Full Path List' column.
    """
    df.loc[:, 'Full Path List'] = df.apply(get_full_path, axis=1).apply(_common.path_string_to_list)


def remove_full_path(df):
    """
    Remove the 'Full Path List' column.
    """
    df.drop('Full Path List', axis=1, inplace=True)


def update_path_from_full_path_list(df):
    """
    From the 'Full Path List' column, set the 'Path' column.
    """
    df['Path'] = df.apply(lambda node: _common.path_list_to_string(node['Full Path List'][:-1]), axis=1)


def trim_unneeded_paths(df, parent_full_path=None, maintain_last_shared_root=None):
    """
    Remove any leading parts of the path that are shared across all rows. Then add the parent_path back onto the
    front of the path.

    E.G. If all rows have a path of 'USA >> Texas >> Houston >> Cooling Tower >> Area {x} >> ...',
    'Cooling Tower' would become the root asset for this Tree. Then if parent_path was 'My Tree >> Cooling Tower',
    all rows would have a path 'My Tree >> Cooling Tower >> Area {x} >> ...'
    """
    if len(df) == 0:
        return df

    # Get the path of the first node. It doesn't matter which we start with since we're only removing paths that are
    # shared across ALL rows.
    decorate_with_full_path(df)
    shared_root = _push.get_common_root(df['Full Path List'])
    # Trim the path until we're left with the last universally shared node.
    while shared_root:
        trimmed_full_path_list = df['Full Path List'].apply(lambda l: l[1:])
        remaining_shared_root = _push.get_common_root(trimmed_full_path_list)
        keep_last_shared_root = True
        if parent_full_path and remaining_shared_root:
            # We only want to remove the root-most path if it is already going to be the parent (due to insert)
            parent_name = _common.path_string_to_list(parent_full_path)[-1]
            keep_last_shared_root = remaining_shared_root != parent_name
        elif parent_full_path and shared_root and isinstance(maintain_last_shared_root, bool):
            # We explicitly want to remove the last shared root so it can be replaced.
            keep_last_shared_root = maintain_last_shared_root
        if not remaining_shared_root and keep_last_shared_root:
            # We need to keep the last shared root so do not save trimmed_full_path_list
            break
        df['Full Path List'] = trimmed_full_path_list
        if 'Depth' in df:
            df['Depth'] -= 1
        shared_root = remaining_shared_root

    if parent_full_path:
        # Prepend the parent path if applicable
        parent_path_list = _common.path_string_to_list(parent_full_path)
        parent_name = parent_path_list[-1]
        if _push.get_common_root(df['Full Path List']) == parent_name:
            parent_path_list.pop()
        if parent_path_list:
            df['Full Path List'] = df['Full Path List'].apply(lambda l: parent_path_list + l)
            if 'Depth' in df:
                df['Depth'] += len(parent_path_list)
    update_path_from_full_path_list(df)
    remove_full_path(df)
    return df


def get_shared_root(full_path_series):
    """
    Returns the highest shared name in the input paths. If no such name exists, returns None
    """
    first_full_path_list = full_path_series.iloc[0]
    if not len(first_full_path_list):
        return None
    root_name = first_full_path_list[0]
    all_roots_same = full_path_series.apply(lambda l: len(l) and l[0] == root_name).all()
    return root_name if all_roots_same else None


def reify_missing_assets(df, existing_parent_path=None):
    """
    Automatically generate any assets that are referred to by path only.
    E.G. If this tree were defined using a dataframe containing only leaf signals, but with a Path column of
    'Cooling Tower >> Area {x} >> {signal}', the 'Cooling Tower' and 'Area {x}' assets would be generated.

    If existing_parent_path is provided, the reification will not occur for any existing parents.
    E.G. 'Example >> Cooling Tower >> Area {x} >> {signal}' with existing_parent_path='Example'
     would only generate 'Cooling Tower' and 'Area {x}' assets, not 'Example'.
    """
    # Store the Full Path tuples of all possible Assets to be created in a set
    full_paths = set()
    for path_list in df.apply(get_full_path, axis=1).apply(_common.path_string_to_list):
        full_paths.update([tuple(path_list[:i]) for i in range(1, len(path_list))])
    # Remove all Assets whose paths are contained in the existing_parent_path
    if existing_parent_path is not None:
        full_paths.difference_update([full_path for full_path in full_paths if
                                      _common.path_list_to_string(full_path) in existing_parent_path])
    # Create dataframe rows based on these paths, and use a single pd.merge call to update the dataframe
    new_assets = pd.DataFrame([{
        'Type': 'Asset',
        'Path': _common.path_list_to_string(full_path[:-1]),
        'Name': full_path[-1],
        'Depth': len(full_path)
    } for full_path in full_paths])
    _utils.drop_duplicate_items(new_assets)
    return _utils.upsert(df, new_assets, prefer_right=False)


def determine_path(row):
    """
    Gets the path from the Path and Asset columns
    """
    path = _common.get(row, 'Path')
    asset = _common.get(row, 'Asset')
    if not isinstance(path, str):
        path = None
    if not isinstance(asset, str):
        asset = None
    return ' >> '.join([s for s in (path, asset) if s is not None])



import concurrent.futures
import datetime
import queue
import re
import threading
from typing import Tuple, Callable, Dict, Optional

import pandas as pd
from IPython.core.display import display, HTML

from seeq.spy import _common
from seeq.spy._errors import *
from seeq.spy._session import Session


class Status:
    RUNNING = 0
    SUCCESS = 1
    FAILURE = 2
    CANCELED = 3

    jobs: Dict[object, Tuple[Tuple, Optional[Callable[[object, object], None]]]]

    def __init__(self, quiet=False):
        """
        A Python Class-based object which tracks the progress status of
        various SPy functions.

        Parameters
        ----------
        quiet : bool, default False
            If True, suppresses progress output. Supersedes the quiet flag
            of any function the status is passed to.
        """
        self.quiet = quiet
        self._df = pd.DataFrame()
        self.timer = _common.timer_start()
        self.message = None
        self.code = None
        self.warnings = set()
        self.printed_warnings = set()
        self.inner = dict()
        self.update_queue = queue.Queue()
        self.interrupted_event = threading.Event()
        self.jobs = dict()
        self.current_df_index = None

    def __str__(self):
        return self.message if self.message else 'Uninitialized'

    def __getstate__(self):
        # We can only pickle certain members. This has to mirror __setstate__().
        return self.quiet, self.df, self.message, self.code, self.warnings, self.inner

    def __setstate__(self, state):
        self.quiet, self.df, self.message, self.code, self.warnings, self.inner = state

    @property
    def df(self):
        _common.clear_properties_on_df(self._df)
        return self._df

    @df.setter
    def df(self, value: pd.DataFrame):
        self._df = value.copy()
        _common.clear_properties_on_df(self._df)

    def create_inner(self, name: str, quiet: bool = None):
        inner_status = Status(quiet=self.quiet if quiet is None else quiet)
        self.inner[name] = inner_status
        return inner_status

    def metrics(self, d):
        self.df = pd.DataFrame(d).transpose()

    def put(self, column, value):
        self.df.at[self.current_df_index, column] = value

    def get(self, column):
        return self.df.at[self.current_df_index, column]

    def warn(self, warning):
        self.warnings.add(warning)

    def _drain_updates(self):
        while True:
            try:
                _index, _updates = self.update_queue.get_nowait()

                for _update_column, _update_value in _updates.items():
                    self.df.at[_index, _update_column] = _update_value

            except queue.Empty:
                break

        self.update()

    def send_update(self, index: object, updates: Dict[str, object]):
        if self.is_interrupted():
            # Raise the exception before we put the update on the queue -- we don't want to incorrectly report success
            raise KeyboardInterrupt()

        self.update_queue.put((index, updates))

    def interrupt(self):
        self.interrupted_event.set()

    def is_interrupted(self):
        return self.interrupted_event.is_set()

    def add_job(self, index: object, func_with_args: Tuple, on_job_success: Callable = None):
        self.jobs[index] = (func_with_args, on_job_success)

    def clear_jobs(self):
        self.jobs = dict()

    def execute_jobs(self, session: Session, errors: str):
        try:
            exception_raised = None
            with concurrent.futures.ThreadPoolExecutor(max_workers=session.options.max_concurrent_requests) as executor:
                _futures = dict()
                for job_index, (func_with_args, on_job_success) in self.jobs.items():
                    _futures[executor.submit(*func_with_args)] = (job_index, on_job_success)

                while True:

                    # noinspection PyBroadException
                    try:
                        self._drain_updates()

                        # Now we wait for all the futures to complete, breaking out every half second to drain status
                        # updates (see TimeoutError except block).
                        for future in concurrent.futures.as_completed(_futures, 0.5):
                            job_index, on_job_success = _futures[future]
                            del _futures[future]
                            self._drain_updates()

                            if future.cancelled() or isinstance(future.exception(), KeyboardInterrupt):
                                self.df.at[job_index, 'Result'] = 'Canceled'
                                continue

                            if future.exception():
                                self.df.at[job_index, 'Result'] = _common.format_exception(future.exception())
                                if errors == 'raise':
                                    raise future.exception()
                                else:
                                    continue

                            if on_job_success:
                                # noinspection PyBroadException
                                try:
                                    on_job_success(job_index, future.result())
                                except BaseException:
                                    self.df.at[job_index, 'Result'] = _common.format_exception()
                                    if errors == 'raise':
                                        raise
                                    else:
                                        continue

                        # We got all the way through the iterator without encountering a TimeoutError, so break
                        break

                    except concurrent.futures.TimeoutError:
                        # Start the loop again from the top, draining the status updates first
                        pass

                    except BaseException as e:
                        for future in _futures.keys():
                            future.cancel()
                        self.interrupt()
                        exception_raised = e

            if exception_raised:
                self.exception(exception_raised, throw=True)

        finally:
            self._drain_updates()
            self.clear_jobs()

    def update(self, new_message=None, new_code=None):
        if self.quiet:
            return

        if new_message is None:
            new_message = self.message

        if new_code is not None:
            self.code = new_code

        if not _common.display_supports_html():
            if new_message != self.message:
                for warning in (self.warnings - self.printed_warnings):
                    display(warning)
                self.printed_warnings = set(self.warnings)

                text = re.sub(r'</?[^>]+>', '', new_message)
                # noinspection PyTypeChecker
                display(text)
                self.message = new_message
            return

        self.message = new_message

        _common.ipython_clear_output(wait=True)

        display_df = self.df
        if self.code == Status.RUNNING and len(self.df) > 20 and 'Result' in self.df.columns:
            display_df = self.df[~self.df['Result'].isin(['Queued', 'Success'])]

        display_df = display_df.head(20)

        if self.code == Status.RUNNING:
            color = '#EEEEFF'
        elif self.code == Status.SUCCESS:
            color = '#EEFFEE'
        else:
            color = '#FFEEEE'

        html = ''
        if len(self.warnings) > 0:
            for warning in self.warnings:
                html += '<div style="background-color: #FFFFCC;">%s</div>' % (Status._massage_cell(warning))

        style = 'background-color: %s;' % color
        html += '<div style="%s">%s</div>' % (
            style + 'text-align: left;', Status._massage_cell(self.message))

        if len(display_df) > 0:
            html += '<table class="tex2jax_ignore">'  # Ignore mathjax renderings so $...$ isn't converted to latex
            html += '<tr><td style="%s"></td>' % style

            for col in display_df.columns:
                align = 'left' if display_df.dtypes[col] == object else 'right'
                html += '<td style="%s text-align: %s;">%s</td>' % (style, align, Status._massage_cell(col))

            html += '</tr>'

            for index, row in display_df.iterrows():
                html += '<tr style="%s">' % style
                html += '<td style="vertical-align: top;">%s</td>' % index
                for cell in row:
                    if isinstance(cell, datetime.timedelta):
                        hours, remainder = divmod(cell.seconds, 3600)
                        minutes, seconds = divmod(remainder, 60)
                        html += '<td style="vertical-align: top;">{:02}:{:02}:{:02}.{:02}</td>'.format(
                            int(hours), int(minutes), int(seconds), int((cell.microseconds + 5000) / 10000))
                    else:
                        align = 'left' if isinstance(cell, str) else 'right'
                        html += '<td style="text-align: %s; vertical-align: top;">%s</td>' % \
                                (align, Status._massage_cell(cell, links=True))
                html += '</tr>'

            html += '</table>'

        # noinspection PyTypeChecker
        display(HTML(html))

    @staticmethod
    def _massage_cell(cell, links=False):
        cell = str(cell)

        def markdown_bullets_to_html_bullets(match):
            lines = [re.sub(r'^- ', '', line) for line in match[1].split('\n')]
            return '<ul><li>%s</li></ul>' % '</li><li>'.join(lines)

        cell = re.sub(r'\n(- .*((\n- .*)+|$))', markdown_bullets_to_html_bullets, cell)
        cell = cell.replace('\n', '<br>')
        if links:
            cell = re.sub(r'(http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+{}]|[!*(), ]|(%[0-9a-fA-F][0-9a-fA-F]))+)',
                          r'<a target="_blank" href="\1">link</a>',
                          cell)

        return cell

    def get_timer(self):
        return _common.timer_elapsed(self.timer)

    def reset_timer(self):
        self.timer = _common.timer_start()

    def exception(self, e, throw=False, use_error_message=False):
        if isinstance(e, KeyboardInterrupt):
            status_message = 'Canceled'
            status_code = Status.CANCELED
        else:
            status_message = 'Error encountered, scroll down to view' if not use_error_message else str(e)
            status_code = Status.FAILURE

        self.update(status_message, status_code)
        if throw:
            raise e

    @staticmethod
    def validate(status, quiet=False):
        """
        :param status: An already-instantiated Status object
        :type status: Status
        :param quiet: If True, suppresses output to Jupyter/console
        :type quiet: bool
        :rtype Status
        :return: The already-instantiated Status object passed in, or a newly-instantiated Status object
        """
        if status is None:
            status = Status(quiet=quiet)

        else:
            if not isinstance(status, Status):
                raise SPyTypeError(f'Argument status must be of type Status, not {type(status)}')
            if quiet:
                status.warn('The quiet flag is ignored when a status object is provided. Instead, use the quiet flag '
                            'of your status object by creating a quiet status with spy.Status(quiet=True), or setting '
                            'status.quiet=True.')
        return status



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="seeq",
    version="56.0.1.183.4",
    author="Seeq Corporation",
    author_email="support@seeq.com",
    description="The Seeq SDK for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.seeq.com",
    packages=setuptools.find_namespace_packages(include=['seeq.*']),
    include_package_data=True,
    install_requires=[
        # These requirements are for seeq.sdk and should match target/python/requirements.txt
        'certifi >= 14.05.14',
        'six >= 1.10',
        'urllib3 >= 1.15.1',
        'requests >= 2.21.0',
        'cryptography >= 3.2',
        # cryptography 3.3 requires pyOpenSSL >= 19.1.0
        'pyOpenSSL >= 19.1.0',

        # These additional requirements are for seeq.spy and should be similar to crab/requirements.prod.txt
        'notebook >= 6.0.0',
        'ipython >= 7.6.1',
        'matplotlib >= 3.1.1',
        'numpy >= 1.16.4',
        'pandas >= 1.0.0',
        'beautifulsoup4 >= 4.8.0',
        'Deprecated >= 1.2.6',
        'Mako >= 1.1.0',
        'Markdown >= 3.3.4',
        'ipywidgets >= 7.5.1',
        'cron-descriptor >= 1.2.24',
        'recurrent >= 0.4.0',
        'dataclasses >= 0.7; python_version == "3.6"'  # This is a polyfill for dataclasses in Python 3.6.
        # They're standard in Python 3.7.
    ],
    classifiers=[
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    zip_safe=False
)



import logging
import os
import sys

import requests

from seeq.spy._session import Session

DEFAULT_REQUESTS_TIMEOUT = 90

# Logger to be used for logging inside executor container
executor_logger = None


def _setup_executor_logging():
    global executor_logger
    log_level = get_log_level_from_executor()

    executor_logger = logging.getLogger("executor_logger")
    exec_handler = logging.StreamHandler(sys.stdout)
    exec_formatter = logging.Formatter('%(levelname)s - %(message)s')

    exec_handler.setFormatter(exec_formatter)
    executor_logger.addHandler(exec_handler)

    executor_logger.setLevel(log_level)

    return executor_logger


def get_executor_logger():
    global executor_logger

    if executor_logger is not None:
        return executor_logger
    else:
        return _setup_executor_logging()


def get_log_level_from_executor():
    return str(os.environ.get('LOG_LEVEL', 'INFO')).upper()


def running_in_datalab():
    return os.environ.get('SEEQ_SDL_CONTAINER_IS_DATALAB') == 'true'


def running_in_executor():
    return os.environ.get('SEEQ_SDL_CONTAINER_IS_EXECUTOR') == 'true'


def get_label_from_executor():
    return os.environ.get('SEEQ_SDL_LABEL') or ''


def get_results_folder():
    return "_Job Results/"


def get_cell_execution_timeout():
    return 86400


def get_execution_notebook():
    return "/seeq/scheduling/ExecutionNotebook.ipynb"


def requests_get(session: Session, url, params=None, timeout=DEFAULT_REQUESTS_TIMEOUT, **kwargs):
    return requests.get(url, params=params, timeout=timeout, verify=session.https_verify_ssl, **kwargs)


def requests_patch(session: Session, url, data=None, timeout=DEFAULT_REQUESTS_TIMEOUT, **kwargs):
    return requests.patch(url, data=data, timeout=timeout, verify=session.https_verify_ssl, **kwargs)


def requests_post(session: Session, url, data=None, json=None, timeout=DEFAULT_REQUESTS_TIMEOUT, **kwargs):
    return requests.post(url, data=data, json=json, timeout=timeout, verify=session.https_verify_ssl, **kwargs)


def requests_put(session: Session, url, data=None, timeout=DEFAULT_REQUESTS_TIMEOUT, **kwargs):
    return requests.put(url, data=data, timeout=timeout, verify=session.https_verify_ssl, **kwargs)
